/* $Id: ccalls.c,v 1.4 2000/11/27 20:09:44 malekith Exp $ */

#include "util.h"

#include <string.h>
#include <setjmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "vm.h"
#include "ccalls.h"
#include "binfmt.h"

#define STACK_SIZE 1000
#define arg(no, type) (type)cpu->sp[-no]

void set_s(struct cpu *cpu, char **dst, const char *src);

i32 arc4rand(i32 x);

static void do_sys_itoa(struct cpu *cpu)
{
	char buf[128];
	
	snprintf(buf, sizeof(buf) - 1, "%d", arg(1, iptr));
	set_s(cpu, &cpu->s, buf);
}

static inline char *strchrx(char *s, int p)
{
	char *r;
	r = strchr(s, p);
	return r ? r + 1 : 0;
}

static void do_sys_rand_string(struct cpu *cpu)
{
	char *s, *p;
	int n = 1;
	
	s = arg(1, char*);

	p = s;
	while ((p = strchrx(p, '|')))
		n++;
		
	n = arc4rand(n);
	
	p = s;
	while (n--)
		p = strchrx(p, '|');
	assert(p);
	
	s = strchr(p, '|');
	n = s ? s - p : strlen(p);

	s = (char*)malloc(n + 1);
	s[n] = 0;
	memcpy(s, p, n);
	set_s(cpu, &cpu->s, s);
	free(s);
}

static void do_sys_dice(struct cpu *cpu)
{
	char *p;
	int a,b,c,sign = 1;
	p = arg(1, char*);
	
	cpu->d = -1;
	
	while (*p && !isdigit(*p))
		p++;
		
	if (*p == 0)
		return;
		
	a = atoi(p);
	while (isdigit(*p))
		p++;
	while (*p && !isdigit(*p))
		p++;
	if (*p == 0) {
		b = a;
		a = 1;
	} else
		b = atoi(p);
		
	while (isdigit(*p))
		p++;
	c = 0;
	while (*p && !isdigit(*p))
		if (*p++ == '-')
			sign = -1;
	if (*p)
		c = sign * atoi(p);

	cpu->d = c + a;
	while (a--)
		cpu->d += arc4rand(b);
}

void do_syscall(struct cpu *cpu)
{

	switch (cpu->d) {
	case sys_exit:	
		cpu->d = arg(1, iptr);
		longjmp(cpu->hlt, OOPS_EXIT);
		break;
	case sys_print:
		fputs(arg(1, char*), stdout);
		break;
	case sys_atoi:
		cpu->d = atoi(arg(1, char*));
		break;
	case sys_itoa:
		do_sys_itoa(cpu);
		break;
	case sys_rand:
		cpu->d = arc4rand(arg(1, iptr));
		break;
	case sys_rand_string:
		do_sys_rand_string(cpu);
		break;
	case sys_dice:
		do_sys_dice(cpu);
		break;
	default:
		longjmp(cpu->hlt, OOPS_BADSYSCALL);
		break;
	}
}
