/* $Id: class.c,v 1.10 2000/11/27 20:09:44 malekith Exp $ */

#include "util.h"
#include "class.h"
#include "y.tab.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

int debug;
struct class_def *system_class;

char *take_string(struct token *t)
{
	char *r;

	r = t->v_string;
	t->v_string = 0;
	return r;
}

static int make_vars(struct class_def *c, struct token *type, struct token *t)
{
	struct var *v;

	if (t == 0)
		return 0;
		
	assert((t->sel == sel_syn && t->v_int == syn_var_names) ||
	       (t->sel == sel_id));
	       
	if (t->sel == sel_syn) {
		return make_vars(c, type, t->left) ||
		       make_vars(c, type, t->right);
	}
		
	v = NEW(struct var);
	v->name = take_string(t);
	v->offset = c->datasize;
	v->ctx = 0;
	if (token2type(&v->type, type))
		return -1;
	c->datasize += size_of(&v->type);
	
	if (debug)
		printf("\t%s %s @%d;\n", name_of(&v->type),
			v->name, v->offset);

	if (ph_find(c->vars, v->name)) {
		oops(t, "redefintion of variable %s in %s", 
		     v->name, c->name);
		return -1;
	}
	
	ph_add(c->vars, v);
	return 0;
}

static int handle_vars(struct class_def *c, struct token *t)
{
	if (t == 0)
		return 0;
		
	if (t->sel != sel_syn)
		halt();
		
	switch (t->v_int) {
	case syn_class_vars:
		return handle_vars(c, t->left) ||
		       handle_vars(c, t->right);
	case syn_var:
		return make_vars(c, t->left, t->right);
	default:
		halt();
	}
}

struct var *make_arg_list(struct token *t, int *sp)
{
	struct var *h, *e, *p;
	
	if (t == 0)
		return 0;
		
	if (t->sel == sel_syn)
		switch (t->v_int) {
		case syn_func_args:
		case syn_proc_args:
			h = make_arg_list(t->left, sp);
			e = make_arg_list(t->right, sp);
			if (h == 0)
				h = e;
			else if (e) {
				for (p = h; p->next; p = p->next)
					/* nothing */;
				p->next = e;
			}
			return h;
		case syn_func_arg:
		case syn_proc_arg:
			h = NEW(struct var);
			h->next = 0;
			token2type(&h->type, t->left);
			*sp += size_of(&h->type);
			h->offset = *sp;
			h->ctx = 0;
			if (t->right) 
				h->name = take_string(t->right);
			else
				h->name = 0;
			return h;
		default:
			halt();
		}
	else
		halt();
}

static int find_id(struct class_def *c, struct class_func *f, struct token *t)
{
	struct class_func *p;
	struct class_def *s;

	if (strcmp(c->name, f->name) == 0) {
		f->id = 0;
		return 0;
	} else if (strcmp(f->name, "die") == 0) {
		f->id = 1;
		return 0;
	}
	
	for (s = c->super; s; s = s->super) 
		if ((p = ph_find(s->funcs, f->name))) {
			f->id = p->id;
			if (proto_args_diffrent(p->args, f->args) ||
			    types_diffrent(&p->type, &f->type)) {
				oops(t, "declaration args of %s::%s() "
				     "doesn't match base %s::%s()",
				     c->name, f->name, s->name,
				     f->name);
				return 1;
			} else 
				goto bye;
		}
		
	f->id = c->funcsize++;
	db_alloc(c->funcs_xrefs, sizeof(u32));
bye:
	((u32*)db_data(c->funcs_xrefs))[f->id] = c->id;
	return 0;
}

static int handle_procs(struct class_def *c, struct token *t)
{
	int sp = 0;
	struct class_func *f;

	if (t == 0)
		return 0;
		
	if (t->sel != sel_syn)
		halt();
		
	switch (t->v_int) {
	case syn_class_procs:
		return  handle_procs(c, t->left) ||
			handle_procs(c, t->right);
		break;
		
	case syn_syscall:
		SYN(t->right, syscall_args);
		assert(t->right->right->sel == sel_int);
	case syn_proc:
		SYN(t->left, proc_sign);
		f = NEW(struct class_func);
		token2type(&f->type, t->left->left);
		assert(t->left->right->sel == sel_id);
		f->name = take_string(t->left->right);
		if (ph_find(c->funcs, f->name)) {
			oops(t, "redefinition of function %s::%s()", 
			     c->name, f->name);
			return -1;
		}
		if (t->v_int == syn_proc) {
			f->args = make_arg_list(t->right, &sp);
			if (find_id(c, f, t))
				return -1;
		} else {
			f->args = make_arg_list(t->right->left, &sp);
			f->id = -1;
			f->syscall = t->right->right->v_int;
		}
		
		ph_add(c->funcs, f);
		if (debug) {
			struct var *a;
			
			printf("\t%s %s(", name_of(&f->type), f->name);
			for (a = f->args; a; a = a->next)
				printf("%s %s @%d%s", name_of(&a->type),
					a->name ? a->name : "",
					a->offset,
					a->next ? ", " : "");
			printf(") @%d;\n", f->id);
		}
		break;
		
	default:
		halt();
	}

	return 0;
}

struct phash *class_hash;
struct class_def *class_head;

int compile_class(struct token *t)
{
	struct class_def *c;
	struct class_func *f;
	struct token *p;
	static int class_id = 0;
	
	if (class_hash == 0)
		class_hash = ph_new();

	c = NEW(struct class_def);
	memset(c, 0, sizeof(*c));

	p = t->left;

	SYN(t, class);
	SYN(p, class_name);
	assert(p->left->sel == sel_id);
	
	c->name = take_string(p->left);
	c->id = class_id++;

	if (debug)
		printf("----> compiling class %s\n", c->name);

	c->funcs_xrefs = db_new();
	
	if (p->right) {
		assert(p->right->sel == sel_id);
		c->super = ph_find(class_hash, p->right->v_string);
	
		if (c->super == 0) {
			oops(p, "parent class %s undefined", 
			     p->right->v_string);
			return 1;
		}

		c->datasize = c->super->datasize;
		c->funcsize = c->super->funcsize;

		db_add(c->funcs_xrefs, c->super->funcs_xrefs, 
		       c->super->funcsize * sizeof(u32));
	} else {
		/* funcID==0 - constructor, 1 - destructor */
		c->funcsize = 2;
		/* first u32 of class is its ID */
		c->datasize = sizeof(u32);
		db_alloc(c->funcs_xrefs, 2 * sizeof(u32));
	}

	/* c/dtor are always defined here */
	((u32*)db_data(c->funcs_xrefs))[0] = c->id;
	((u32*)db_data(c->funcs_xrefs))[1] = c->id;
	
	p = t->right;
	SYN(p, class_content);

	c->vars = ph_new();
	c->funcs = ph_new();
	
	ph_add(class_hash, c);

	if (handle_vars(c, p->left))
		return 1;
	
	if (debug)	
		printf("proc:\n");
	
	if (handle_procs(c, p->right))
		return 1;

	f = (struct class_func*)ph_find(c->funcs, c->name);
	if (f == 0) {
		oops(t, "constructor %s::%s() not declared", 
		     c->name, c->name);
		return 1;
	}
	if (f->args) {
		oops(t, "constructor %s::%s() has args", 
		     c->name, c->name);
		return 1;
	}

	f = (struct class_func*)ph_find(c->funcs, "die");
	c->def_dtor = (f == 0);
	if (f && f->args) {
		oops(t, "destructor %s::die() has args", c->name);
		return 1;
	}
	
	if (c->def_dtor) {
		f = NEW(struct class_func);
		f->type.id = sel_void;
		f->type.ptr = 0;
		f->name = strdup("die");
		f->args = 0;
		f->id = 1;
		ph_add(c->funcs, f);
	}
	
	if (debug)
		printf("data: %d\nfunc: %d\n\n", c->datasize, c->funcsize);

	c->lnext = 0;
	
	if (class_head == 0)
		class_head = c;
	else {
		struct class_def *p;

		for (p = class_head; p->lnext; p = p->lnext)
			/* nothing */;
		p->lnext = c;
	}

	c->funcs_xrefs = db_finish(c->funcs_xrefs, 0);
	
	if (strcmp(c->name, "Sys") == 0)
		system_class = c;
		
	destroy_token(t);
	return 0;
}
